<?php

defined ('_JEXEC') or die('Restricted access');

/**
 * Shipment plugin for general, rules-based shipments, like regular postal services with complex shipping cost structures
 * Advanced part, implementing general mathematical expression evaluation
 *
 * @version $Id$
 * @package VirtueMart
 * @subpackage Plugins - shipment.
 * @copyright Copyright (C) 2013 Reinhold Kainhofer, reinhold@kainhofer.com
 * @license http://www.gnu.org/copyleft/gpl.html GNU/GPL, see LICENSE.php
 * VirtueMart is free software. This version may have been modified pursuant
 * to the GNU General Public License, and as distributed it includes or
 * is derivative of works licensed under the GNU General Public License or
 * other free or open source software licenses.
 * See /administrator/components/com_virtuemart/COPYRIGHT.php for copyright notices and details.
 *
 * http://virtuemart.org
 * @author Reinhold Kainhofer, based on the weight_countries shipping plugin by Valerie Isaksen
 *
 */

if (!class_exists ('vmPSPlugin')) {
	require(JPATH_VM_PLUGINS . DS . 'vmpsplugin.php');
}
if (!class_exists ('plgVmShipmentRules_Shipping')) {
	require (dirname(__FILE__).DS.'rules_shipping.php');
}

/** Shipping costs according to general rules.
 *  Derived from the standard plugin, no need to change anything! The standard plugin already uses the advanced rules class defined below, if it can be found
 */
class plgVmShipmentRules_Shipping_Advanced extends plgVmShipmentRules_Shipping {
// 	function __construct (& $subject, $config) {
// 		parent::__construct ($subject, $config);
// 	}

}
 
/** Extend the shipping rules by allowing arbitrary mathematical expressions
 */
class ShippingRule_Advanced extends ShippingRule {
	var $operators = array("+"=>10, "-"=>10, "*"=>20, "/"=>20, "%"=>30, "^"=>50, "("=>100, ")"=>100);

	function __construct ($rule, $countries, $tax_id) {
		parent::__construct ($rule, $countries, $tax_id);
	}

	/** parse the mathematical expressions (following Knuth 1962):
	 *  First parse the string into an array of tokens (operators and operands) by a simple regexp with known operators as separators)
	 *  Then convert the infix notation into postfix (RPN), taking care of operator precedence
	 *    1) Initialize empty stack and empty result variable
	 *    2) Read infix expression from left to right, one atom at a time
	 *    3) If operand => Append to result
	 *    4) If operator:
	 *        4a) Pop operators from stack until opening parenthesis, operator of 
	 *            lower precedence or right-associative symbol of equal precedence.
	 *        4b) Push operator onto stack
	 *    5) If opening parenthesis => push onto stack
	 *    6) If closing parenthesis:
	 *        6a) Pop operators from stack until opening parenthesis is found
	 *        6b) push them to the result (not the opening parenthesis, of course)
	 *    7) At the end of the input, pop all operators from the stack and onto the result
	 *  Store this RPN list for later evaluation
	 */
	function parseShippingTerm ($expr) {
		$op_re = ':\s*(\+|-|\*|/|%|\(|\)|\^)\s*:';
		$atoms = preg_split($op_re, strtolower($expr), -1, PREG_SPLIT_DELIM_CAPTURE|PREG_SPLIT_NO_EMPTY);
		if (count($atoms)==1) return $atoms[0];
		// Operators,including precedence
		
		$stack = array ();  // 1)
		$result = array ();
		foreach ($atoms as $a) { // 2)
			if (!isset($this->operators[$a])) { // 3)
				array_push ($result, $a);
			} elseif ($a == "(") { // 5)
				array_push ($stack, $a);
			} elseif ($a == ")") { // 6)
				do {
					$op=array_pop($stack); // 6a)
					if (!empty($op) && $op != "(") {
						array_push ($result, $op); // 6b)
					} else {
						if ($op != "(") {
							// If no ( can be found, the expression is wrong!
							JFactory::getApplication()->enqueueMessage(JText::sprintf('VMSHIPMENT_RULES_PARSE_MISSING_PAREN', $expr), 'error');
						}
						break; // We have found the opening parenthesis
					}
				} while (0);
			} else { // 4
				// For operators, pop operators from the stack until you reach an opening parenthesis, 
				// an operator of lower precedence, or a right associative symbol of equal precedence. 
				$prec = $this->operators[$a];
				
				while (count($stack)>0) { // 4a)
					$op = array_pop ($stack);
					// Ignore the right-associative symbols of equal precedence for now...
					if ($op == "(") {
						break;
					} elseif ($this->operators[$op]<$prec) {
						// We found an operator with lower precedence, add it back to the stack!
						array_push ($stack, $op); // 4b)
						break;
					} else {
						array_push ($result, $op);
					}
				} while (0);
				array_push ($stack, $a); // 4b)
			}
		}
		// Finally, pop all operators from the stack and append them to the result
		while ($op=array_pop($stack)) {
			// Opening parentheses should not be found on the stack any more. That would mean a closing paren is missing!
			if ($op == "(") {
				JFactory::getApplication()->enqueueMessage(JText::sprintf('VMSHIPMENT_RULES_PARSE_PAREN_NOT_CLOSED', $expr), 'error');
			} else {
				array_push ($result, $op);
			}
		}
		/* In the advanced version, shipping cost can be given as a full mathematical expression */
		return $result;
	}

	function evaluateTerm ($expr, $vals) {
		if (!is_array($expr)) {
			// Normal expression, no complex mathematical formula in RPN
			if (!is_numeric($expr) && isset($vals[$expr])) $expr=$vals[$expr];
			if (!is_numeric($expr)) {
				JFactory::getApplication()->enqueueMessage(JText::sprintf('VMSHIPMENT_RULES_EVALUATE_NONNUMERIC', $expr, $this->rulestring), 'error');
				return 0;
			}
			return $expr;
		}

		/** Evaluate the RPN, according to Knuth: 
		 *   1) Initialize an empty stack
		 *   2) Read the RPN from left to right
		 *   3) If operand, push it onto the stack
		 *   4) If operator:
		 *       4a) pop two operands
		 *       4b) perform operation
		 *       4c) push result onto stack
		 *       4d) (If less than two operands => ERROR, invalid syntax)
		 *   5) At the end of the RPN, pop the result from the stack. 
		 *       5a) The stack should now be empty (otherwise, ERROR, invalid syntax)
		 */
		$stack=array(); // 1)
		foreach ($expr as $e) { // 2)
			if (!isset($this->operators[$e])) { // 3)
				// Operand => push onto stack
				if (!is_numeric($e) && isset($vals[$e])) $e = $vals[$e];
				if (!is_numeric($e)) {
					JFactory::getApplication()->enqueueMessage(JText::sprintf('VMSHIPMENT_RULES_EVALUATE_NONNUMERIC', $e, $this->rulestring), 'error');
					$e = 0;
				}
				array_push ($stack, $e);
			} else { // 4)
				// Operator => apply to the last two values on the stack
				if (count($stack)<2) { // 4d)
					JFactory::getApplication()->enqueueMessage(JText::sprintf('VMSHIPMENT_RULES_EVALUATE_SYNTAXERROR', $this->rulestring), 'error');
					array_push($stack, 0);
					continue;
				}
				$o2 = array_pop($stack); // 4a)
				$o1 = array_pop($stack);
				$res = 0;
				switch ($e) { // 4b)
					case "+": $res = $o1+$o2; break;
					case "-": $res = $o1-$o2; break;
					case "*": $res = $o1*$o2; break;
					case "/": $res = $o1/$o2; break;
					case "%": $res = $o1%$o2; break;
					case "^": $res = $o1^$o1; break;
					default: 
						JFactory::getApplication()->enqueueMessage(JText::sprintf('VMSHIPMENT_RULES_EVALUATE_UNKNOWN_OPERATOR', $e, $this->rulestring), 'error');
						$res = 0;
				}
				array_push($stack, $res); // 4c)
			}
		}
		// 5a)
		if (count($stack) != 1) {
			JFactory::getApplication()->enqueueMessage(JText::sprintf('VMSHIPMENT_RULES_EVALUATE_UNKNOWN_ERROR', $this->rulestring), 'error');
			$stack = array (0);
		}
		$res = array_pop($stack); // 5)
		return $res;
	}

}

// No closing tag
